import React, {Component} from 'react'
import PropTypes from 'prop-types'
import { is, fromJS } from 'immutable'
import { Table, Typography, Modal, Input, InputNumber, Switch, Button, notification, message, Select, DatePicker } from 'antd'
import { EditOutlined, QuestionCircleOutlined } from '@ant-design/icons'
import moment from 'moment'
import md5 from 'md5'

import Api from '@/api'
import asyncComponent from '@/utils/asyncComponent'
import Utils, { getEditTableSql, getMark } from '@/utils/utils.js'
import MKEmitter from '@/utils/events.js'
import Encrypts from '@/components/encrypts'
import './index.scss'

const { Paragraph } = Typography
const MkIcon = asyncComponent(() => import('@/components/mk-icon'))
const MKPopSelect = asyncComponent(() => import('./mkPopSelect'))
const CardCellComponent = asyncComponent(() => import('@/tabviews/custom/components/card/cardcellList'))

class MkSwitch extends Component {
  static propTpyes = {
    autoFocus: PropTypes.bool,
    defaultValue: PropTypes.any,
    config: PropTypes.object,
    onChange: PropTypes.func
  }

  state = {
    status: false
  }

  node = null

  UNSAFE_componentWillMount () {
    const { defaultValue, config } = this.props
    
    let status = false

    if (defaultValue === config.openVal) {
      status = true
    }
    
    this.setState({status})
  }

  changeStatus = (val) => {
    const { config, lineId } = this.props

    this.setState({ status: val })

    let values = {[config.field]: val ? config.openVal : config.closeVal}
    
    this.props.onChange(values, '')

    this.node.blur()
    
    if (/\$noAct/.test(config.enter)) return

    if (/\$next/.test(config.enter)) {
      MKEmitter.emit('nextLine' + config.tableId, lineId, config.enter.replace('$next_', ''))
    } else {
      MKEmitter.emit('setFocus' + config.tableId, lineId, config.enter)
    }
  }

  onFocus = () => {
    const { config, lineId } = this.props

    if (!config.$ctrl) return

    MKEmitter.emit('colFocus' + config.tableId, lineId, config.uuid)
  }

  onBlur = () => {
    const { config, lineId } = this.props

    if (config.$ctrl) {
      MKEmitter.emit('colBlur' + config.tableId, lineId, config.uuid)
    }

    this.props.onBlur && this.props.onBlur()
  }

  render() {
    const { config, autoFocus } = this.props
    const { status } = this.state

    return (
      <Switch
        ref={ref => this.node = ref}
        checkedChildren={config.openText}
        checked={status}
        autoFocus={autoFocus}
        onFocus={this.onFocus}
        onBlur={this.onBlur}
        unCheckedChildren={config.closeText}
        onChange={this.changeStatus}
      />
    )
  }
}

class MkDatePicker extends Component {
  static propTpyes = {
    defaultValue: PropTypes.any,
    config: PropTypes.object,
    onChange: PropTypes.func
  }
  
  state = {
    value: null,
    open: false
  }

  UNSAFE_componentWillMount () {
    const { defaultValue } = this.props
    
    let _value = defaultValue || null
    if (_value) {
      _value = moment(_value, 'YYYY-MM-DD HH:mm:ss')
    }
    
    this.setState({value: _value})
  }

  componentDidMount() {
    const { config, autoFocus, lineId } = this.props

    if (autoFocus === true) {
      this.setState({open: true})

      if (config.$ctrl) {
        MKEmitter.emit('colFocus' + config.tableId, lineId, config.uuid)
      }
    }

    MKEmitter.addListener('setFocus' + config.tableId, this.setFocus)
  }

  componentWillUnmount () {
    this.setState = () => {
      return
    }

    MKEmitter.removeListener('setFocus' + this.props.config.tableId, this.setFocus)
  }

  setFocus = (lId, colId) => {
    const { config, lineId } = this.props

    if (lId !== lineId || config.uuid !== colId) return

    this.setState({open: true})

    if (config.$ctrl) {
      MKEmitter.emit('colFocus' + config.tableId, lineId, config.uuid)
    }
  }

  onOpenChange = (open) => {
    const { config, lineId } = this.props

    this.setState({open})

    if (!open) {
      this.props.onBlur && this.props.onBlur()
    }

    if (!config.$ctrl) return

    if (open) {
      MKEmitter.emit('colFocus' + config.tableId, lineId, config.uuid)
    } else {
      MKEmitter.emit('colBlur' + config.tableId, lineId, config.uuid)
    }
  }

  onChange = (val) => {
    const { config, lineId } = this.props

    let _val = val ? moment(val).format(config.format) : ''

    if (config.precision === 'hour') {
      _val = _val + ':00:00'
    } else if (config.precision === 'minute') {
      _val = _val + ':00'
    }

    let values = {[config.field]: _val}

    this.props.onChange(values)

    if (/\$noAct/.test(config.enter)) return

    if (/\$next/.test(config.enter)) {
      MKEmitter.emit('nextLine' + config.tableId, lineId, config.enter.replace('$next_', ''))
    } else {
      MKEmitter.emit('setFocus' + config.tableId, lineId, config.enter)
    }
  }

  render() {
    const { config } = this.props
    const { value, open } = this.state

    return (
      <DatePicker dropdownClassName={'mk-date-picker ' + config.precision} showTime={config.format !== 'YYYY-MM-DD'} format={config.format} open={open} defaultValue={value} onChange={this.onChange} onOpenChange={this.onOpenChange}/>
    )
  }
}

class MkSelect extends Component {
  static propTpyes = {
    defaultValue: PropTypes.any,
    lineId: PropTypes.string,
    config: PropTypes.object,
    onChange: PropTypes.func,
  }
  
  state = {
    value: ''
  }

  node = null

  UNSAFE_componentWillMount() {
    const { defaultValue } = this.props

    this.setState({value: defaultValue})
  }

  componentDidMount() {
    const { config, autoFocus, lineId } = this.props

    if (autoFocus) {
      let node = document.getElementById(config.uuid + lineId)
      node && node.click()
    }
    MKEmitter.addListener('setFocus' + config.tableId, this.setFocus)
  }

  UNSAFE_componentWillReceiveProps(nextProps) {
    if (nextProps.defaultValue !== this.props.defaultValue) {
      this.setState({value: nextProps.defaultValue})
    }
  }

  componentWillUnmount () {
    this.setState = () => {
      return
    }

    MKEmitter.removeListener('setFocus' + this.props.config.tableId, this.setFocus)
  }

  setFocus = (lId, colId) => {
    const { config, lineId } = this.props

    if (lId !== lineId || config.uuid !== colId) return

    let node = document.getElementById(config.uuid + lineId)
    node && node.click()
  }

  onSelectChange = (val, option) => {
    const { config, lineId } = this.props

    let values = {}
    let _option = config.options.filter(m => m.key === option.key)[0]

    if (_option) {
      if (config.linkSubField) {
        config.linkSubField.forEach((m, i) => {
          values[m] = _option[m] !== undefined ? _option[m] : ''
        })
      }

      values[config.field] = val
    }

    this.setState({value: val})

    this.props.onChange(values, val)

    if (config.enter === '$noActX') return

    this.node.blur()
    
    if (/\$noAct/.test(config.enter)) return

    setTimeout(() => {
      if (/\$next/.test(config.enter)) {
        MKEmitter.emit('nextLine' + config.tableId, lineId, config.enter.replace('$next_', ''))
      } else {
        MKEmitter.emit('setFocus' + config.tableId, lineId, config.enter)
      }
    }, 100)
  }

  onFocus = () => {
    const { config, lineId } = this.props

    if (!config.$ctrl) return

    MKEmitter.emit('colFocus' + config.tableId, lineId, config.uuid)
  }

  onBlur = () => {
    const { config, lineId } = this.props

    if (config.$ctrl) {
      MKEmitter.emit('colBlur' + config.tableId, lineId, config.uuid, true)
    }

    setTimeout(() => {
      this.props.onBlur && this.props.onBlur()
    }, 10)
  }

  render() {
    const { config, lineId } = this.props
    const { value } = this.state

    return (
      <Select
        showSearch
        dropdownClassName="edit-table-dropdown"
        dropdownMatchSelectWidth={config.dropdown === 'fixed'}
        value={value}
        id={config.uuid + lineId}
        ref={ref => this.node = ref}
        onFocus={this.onFocus}
        onBlur={this.onBlur}
        filterOption={(input, option) => option.props.children.toLowerCase().indexOf(input.toLowerCase()) >= 0}
        onSelect={this.onSelectChange}
      >
        {config.options.map(item => (<Select.Option key={item.key} disabled={item.$disabled} value={item.value}>{item.label}</Select.Option>))}
      </Select>
    )
  }
}

class MkInput extends Component {
  static propTpyes = {
    defaultValue: PropTypes.any,
    lineId: PropTypes.string,
    config: PropTypes.object,
    onChange: PropTypes.func
  }
  
  state = {
    value: null,
    err: null
  }

  node = null

  UNSAFE_componentWillMount() {
    const { defaultValue } = this.props

    this.setState({value: defaultValue})
  }

  componentDidMount() {
    const { config, autoFocus } = this.props

    if (autoFocus) {
      this.node.select()
    }
    MKEmitter.addListener('setFocus' + config.tableId, this.setFocus)
  }

  UNSAFE_componentWillReceiveProps(nextProps) {
    if (nextProps.defaultValue !== this.props.defaultValue) {
      this.setState({value: nextProps.defaultValue})
    }
  }

  componentWillUnmount () {
    this.setState = () => {
      return
    }

    MKEmitter.removeListener('setFocus' + this.props.config.tableId, this.setFocus)
  }

  setFocus = (lId, colId) => {
    const { config, lineId } = this.props

    if (lId !== lineId || config.uuid !== colId) return

    this.node.select()
  }

  onChange = (val) => {
    this.setState({value: val})
  }

  onFocus = () => {
    const { config, lineId } = this.props

    if (!config.$ctrl) return

    MKEmitter.emit('colFocus' + config.tableId, lineId, config.uuid)
  }

  enterPress = () => {
    const { config, lineId } = this.props

    this.node.blur()
    
    if (/\$noAct/.test(config.enter)) return

    if (/\$next/.test(config.enter)) {
      MKEmitter.emit('nextLine' + config.tableId, lineId, config.enter.replace('$next_', ''))
    } else {
      MKEmitter.emit('setFocus' + config.tableId, lineId, config.enter)
    }
  }

  onBlur = () => {
    const { config, lineId } = this.props
    const { value } = this.state
    
    let err = null

    if (config.required === 'true' && !value) {
      err = '请填写' + config.label
    }

    this.setState({err})

    this.props.onChange({[config.field]: value})

    if (config.$ctrl) {
      MKEmitter.emit('colBlur' + config.tableId, lineId, config.uuid)
    }

    this.props.onBlur && this.props.onBlur()
  }

  render() {
    const { value, err } = this.state

    return (
      <Input
        title={err}
        className={err ? 'has-error' : ''}
        ref={ref => this.node = ref}
        value={value}
        onChange={(e) => this.onChange(e.target.value)}
        onPressEnter={this.enterPress}
        onFocus={this.onFocus}
        onBlur={this.onBlur}
      />
    )
  }
}

class MkInputNumber extends Component {
  static propTpyes = {
    defaultValue: PropTypes.any,
    lineId: PropTypes.string,
    config: PropTypes.object,
    onChange: PropTypes.func
  }
  
  state = {
    value: null,
    err: null,
    clear: false
  }

  node = null

  UNSAFE_componentWillMount() {
    const { defaultValue } = this.props

    this.setState({value: defaultValue})
  }

  componentDidMount() {
    const { config, autoFocus } = this.props

    if (autoFocus) {
      this.node.focus()
    }
    MKEmitter.addListener('setFocus' + config.tableId, this.setFocus)
  }

  UNSAFE_componentWillReceiveProps(nextProps) {
    if (nextProps.defaultValue !== this.props.defaultValue) {
      this.setState({value: nextProps.defaultValue})
    }
  }

  componentWillUnmount () {
    this.setState = () => {
      return
    }

    MKEmitter.removeListener('setFocus' + this.props.config.tableId, this.setFocus)
  }

  setFocus = (lId, colId) => {
    const { config, lineId } = this.props

    if (lId !== lineId || config.uuid !== colId) return

    this.node.focus()
  }

  onChange = (val) => {
    const { config } = this.props

    this.setState({value: val})

    if (config.clearField && val && !this.state.clear) {
      this.setState({clear: true})

      this.props.onChange({[config.clearField]: ''})
    }
  }

  onFocus = () => {
    const { config, lineId } = this.props

    this.setState({clear: false})

    if (!config.$ctrl) return

    MKEmitter.emit('colFocus' + config.tableId, lineId, config.uuid)
  }

  enterPress = () => {
    const { config, lineId } = this.props

    this.node.blur()
    
    if (/\$noAct/.test(config.enter)) return

    setTimeout(() => {
      if (/\$next/.test(config.enter)) {
        MKEmitter.emit('nextLine' + config.tableId, lineId, config.enter.replace('$next_', ''))
      } else {
        MKEmitter.emit('setFocus' + config.tableId, lineId, config.enter)
      }
    }, 10)
  }

  onBlur = () => {
    const { config, lineId } = this.props

    if (config.$ctrl) {
      MKEmitter.emit('colBlur' + config.tableId, lineId, config.uuid)
    }

    setTimeout(() => {
      let err = null
      let value = this.state.value

      if (config.noValue === 'hide' && !value) {
        value = 0
      }

      if (config.required === 'true' && !value) {
        err = `${config.label}不可为${config.noValue === 'hide' ? '空' : '0'}`
      } else {
        if (typeof(config.max) === 'number' && value > config.max) {
          err = config.label + '最大为' + config.max
        } else if (typeof(config.min) === 'number' && value < config.min) {
          err = config.label + '最小为' + config.min
        }
      }

      this.setState({err})

      this.props.onChange({[config.field]: value})

      this.props.onBlur && this.props.onBlur()
    }, 5)
  }

  render() {
    const { config } = this.props
    const { value, err } = this.state

    return (
      <InputNumber
        title={err}
        className={err ? 'has-error' : ''}
        ref={ref => this.node = ref}
        precision={config.decimal || 0}
        value={value}
        onChange={(value) => this.onChange(value)}
        onPressEnter={this.enterPress}
        onFocus={this.onFocus}
        onBlur={this.onBlur}
      />
    )
  }
}

class BodyRow extends React.Component {
  shouldComponentUpdate (nextProps, nextState) {
    return !is(fromJS(this.props.data), fromJS(nextProps.data))
  }

  render() {
    let { lineMarks, data, ...resProps } = this.props
    let style = {}
    let className = ''

    lineMarks && lineMarks.some(mark => {
      let originVal = data[mark.field[0]] + ''
      let contrastVal = ''
      let result = false

      if (mark.field[1] === 'static') {
        contrastVal = mark.contrastValue + ''
      } else {
        contrastVal = data[mark.field[2]] + ''
      }

      if (mark.match === '=') {
        result = originVal === contrastVal
      } else if (mark.match === '!=') {
        result = originVal !== contrastVal
      } else if (mark.match === 'like') {
        result = originVal.indexOf(contrastVal) > -1
      } else if (mark.match === '>') {
        try {
          originVal = parseFloat(originVal)
          contrastVal = parseFloat(contrastVal)
        } catch (e) {
          originVal = NaN
        }

        if (!isNaN(originVal) && !isNaN(contrastVal) && originVal > contrastVal) {
          result = true
        }
      } else if (mark.match === '<') {
        try {
          originVal = parseFloat(originVal)
          contrastVal = parseFloat(contrastVal)
        } catch (e) {
          originVal = NaN
        }

        if (!isNaN(originVal) && !isNaN(contrastVal) && originVal < contrastVal) {
          result = true
        }
      }

      if (result) {
        if (mark.signType[0] === 'font') {
          style.color = mark.color
        } else if (mark.signType[0] === 'background') {
          style.background = mark.color
          if (mark.fontColor) {
            style.color = mark.fontColor
          }
          className = 'background'
        } else if (mark.signType[0] === 'underline') {
          style.textDecoration = 'underline'
          style.color = mark.color
        } else if (mark.signType[0] === 'line-through') {
          style.textDecoration = 'line-through'
          style.color = mark.color
        }
      }

      return result
    })

    return <tr {...resProps} className={className} style={style}/>
  }
}

class BodyCell extends React.Component {
  state = {
    editing: false
  }

  componentDidMount() {
    const { col } = this.props

    if (col && col.editable === 'true') {
      MKEmitter.addListener('setFocus' + col.tableId, this.setFocus)
    }
  }

  shouldComponentUpdate (nextProps, nextState) {
    return !is(fromJS(this.props.record), fromJS(nextProps.record)) || !is(fromJS(this.state), fromJS(nextState))
  }

  /**
   * @description 组件销毁，清除state更新，清除快捷键设置
   */
  componentWillUnmount () {
    this.setState = () => {
      return
    }
    if (this.props.col) {
      MKEmitter.removeListener('setFocus' + this.props.col.tableId, this.setFocus)
    }
  }

  setFocus = (lId, colId) => {
    const { col, record } = this.props

    if (lId !== record.$$uuid || col.uuid !== colId) return
    if (col.ctrlField && col.ctrlValue.includes(record[col.ctrlField] + '')) return

    this.setState({editing: true})
  }

  focus = () => {
    const { col, record } = this.props

    if (col.ctrlField && col.ctrlValue.includes(record[col.ctrlField] + '')) return

    this.setState({editing: true})
  }

  onColChange = (values) => {
    const { col, record } = this.props

    MKEmitter.emit('changeRecord' + col.tableId, {...record, ...values})
  }

  render() {
    let { col, record, style, className, ...resProps } = this.props
    const { editing } = this.state

    if (!col) return (<td {...resProps} className={className} style={style}/>)

    let disabled = false
    if (col.ctrlField) {
      disabled = col.ctrlValue.includes(record[col.ctrlField] + '')
    }

    let children = null
    if (col.type === 'text') {
      let content = ''
      if (record[col.field] !== undefined) {
        content = `${record[col.field]}`
      }

      if (col.editType === 'select' && col.options.length > 0) {
        content = col.map.get(content) || content
      } else if (col.editType === 'switch') {
        if (content === col.openVal) {
          content = col.openText
        } else if (content === col.closeVal) {
          content = col.closeText
        }
      } else if (col.editType === 'popSelect') {
        if (col.showField) {
          content = record[col.showField] || content
        }
      }

      if (content !== '') {
        if (col.textFormat === 'YYYY-MM-DD' && /^[1-9]\d{3}(-|\/)(0[1-9]|1[0-2])(-|\/)(0[1-9]|[1-2][0-9]|3[0-1])/.test(content)) {
          content = `${content.substr(0, 4)}-${content.substr(5, 2)}-${content.substr(8, 2)}`
        } else if (col.textFormat === 'YYYY-MM-DD HH:mm:ss' && /^[1-9]\d{3}(-|\/)(0[1-9]|1[0-2])(-|\/)(0[1-9]|[1-2][0-9]|3[0-1]).([0-1][0-9]|2[0-3]):[0-5][0-9]:[0-5][0-9]/.test(content)) {
          content = `${content.substr(0, 4)}-${content.substr(5, 2)}-${content.substr(8, 2)} ${content.substr(11, 2)}:${content.substr(14, 2)}:${content.substr(17, 2)}`
        } else if (col.textFormat === 'encryption') {
          content = <span>{col.prefix || ''}<Encrypts value={content} />{col.postfix || ''}</span>
        }

        if (col.noValue === 'hide' && content < '1949-10-02') {
          content = ''
        }

        if (col.textFormat !== 'encryption') {
          content = (col.prefix || '') + content + (col.postfix || '')
        }
      }

      if (col.marks) {
        style = style ? {...style} : {}

        let mark = getMark(col.marks, record, style)

        if (mark.icon) {
          if (mark.position === 'front') {
            content = <span><MkIcon style={mark.innerStyle} type={mark.icon} /> {content}</span>
          } else {
            content = <span>{content} <MkIcon style={mark.innerStyle} type={mark.icon} /></span>
          }
        } else if (mark.innerStyle) {
          content = <span style={mark.innerStyle}>{content}</span>
        } else if (mark.space) {
          content = <><span dangerouslySetInnerHTML={{__html: mark.space}}></span>{content}</>
        } else if (mark.point) {
          if (mark.position === 'front') {
            content = <>{mark.point}{content}</>
          } else {
            content = <>{content}{mark.point}</>
          }
        }
      }

      if (col.editable === 'true' && !disabled) {
        if (editing) {
          let _value = record[col.field] !== undefined ? record[col.field] : ''

          if (!col.editType || col.editType === 'text') {
            return (<td onClick={(e) => e.stopPropagation()} className="editing_table_cell">
              <MkInput config={col} lineId={record.$$uuid} defaultValue={_value} autoFocus={true} onChange={this.onColChange} onBlur={() => this.setState({editing: false})}/>
            </td>)
          } else if (col.editType === 'switch') {
            return (<td onClick={(e) => e.stopPropagation()} className="editing_table_cell">
              <MkSwitch config={col} lineId={record.$$uuid} defaultValue={_value} autoFocus={true} onChange={this.onColChange} onBlur={() => this.setState({editing: false})}/>
            </td>)
          } else if (col.editType === 'date') {
            return (<td onClick={(e) => e.stopPropagation()} className="editing_table_cell">
              <MkDatePicker config={col} lineId={record.$$uuid} defaultValue={_value || null} autoFocus={true} onChange={this.onColChange} onBlur={() => this.setState({editing: false})}/>
            </td>)
          } else if (col.editType === 'popSelect') {
            let showValue = ''
            if (col.showField) {
              showValue = record[col.showField] || ''
            }

            return (<td onClick={(e) => e.stopPropagation()} className="editing_table_cell">
              <MKPopSelect config={col} lineId={record.$$uuid} defaultValue={_value} showValue={showValue} BID={record.$$BID} autoFocus={true} onChange={this.onColChange} onBlur={() => this.setState({editing: false})}/>
            </td>)
          } else {
            return (<td onClick={(e) => e.stopPropagation()} className="editing_table_cell">
              <MkSelect config={col} lineId={record.$$uuid} defaultValue={_value} autoFocus={true} onChange={this.onColChange} onBlur={() => this.setState({editing: false})}/>
            </td>)
          }
        } else {
          return (<td onClick={(e) => e.stopPropagation()} className={className + ' pointer'} style={style}>
            <div className="mk-mask" id={col.uuid + record.$$uuid} onClick={this.focus}></div>{content}
          </td>)
        }
      } else {
        children = content
      }
    } else if (col.type === 'number') {
      let content = ''
      try {
        content = parseFloat(record[col.field])
        if (isNaN(content)) {
          content = ''
        }
        if (col.noValue === 'hide' && content === 0) {
          content = ''
        }
      } catch (e) {
        content = ''
      }

      if (content !== '') {
        if (col.round) {
          content = Math.round(content * col.round) / col.round
        }
        if (col.format === 'percent') {
          content = content * 100
        } else if (col.format === 'abs') {
          content = Math.abs(content)
        }
        if (col.round) {
          content = content.toFixed(col.decimal)
        }
  
        if (col.format === 'thdSeparator') {
          content = content + ''
          content = content.replace(/\d{1,3}(?=(\d{3})+(\.\d*)?$)/g, '$&,')
        }
  
        content = (col.prefix || '') + content + (col.postfix || '')
      }

      if (col.marks) {
        style = style ? {...style} : {}

        let mark = getMark(col.marks, record, style)

        if (mark.icon) {
          if (mark.position === 'front') {
            content = <span><MkIcon style={mark.innerStyle} type={mark.icon} /> {content}</span>
          } else {
            content = <span>{content} <MkIcon style={mark.innerStyle} type={mark.icon} /></span>
          }
        } else if (mark.innerStyle) {
          content = <span style={mark.innerStyle}>{content}</span>
        } else if (mark.space) {
          content = <><span dangerouslySetInnerHTML={{__html: mark.space}}></span>{content}</>
        } else if (mark.point) {
          if (mark.position === 'front') {
            content = <>{mark.point}{content}</>
          } else {
            content = <>{content}{mark.point}</>
          }
        }
      }

      if (col.editable === 'true' && !disabled) {
        if (editing) {
          let val = record[col.field] !== undefined ? record[col.field] : ''
          if (col.noValue === 'hide' && val === 0) {
            val = ''
          }
          return (<td onClick={(e) => e.stopPropagation()} className="editing_table_cell">
            <MkInputNumber config={col} lineId={record.$$uuid} defaultValue={val} autoFocus={true} onChange={this.onColChange} onBlur={() => this.setState({editing: false})}/>
          </td>)
        } else {
          return (<td onClick={(e) => e.stopPropagation()} className={className + ' pointer'} style={style}>
            <div className="mk-mask" id={col.uuid + record.$$uuid} onClick={this.focus}></div>{content}
          </td>)
        }
      } else {
        children = content
      }
    } else if (col.type === 'textarea') {
      let content = ''
      if (record[col.field] !== undefined) {
        content = `${record[col.field]}`
      }

      if (content) {
        content = (col.prefix || '') + content + (col.postfix || '')
      }

      children = (
        <div>
          {content ? <Paragraph copyable ellipsis={{ rows: 3, expandable: true }}>{content}</Paragraph> : null }
        </div>
      )
    } else if (col.type === 'formula') {
      let content = col.formula

      if (col.eval === 'func') {
        try {
          // eslint-disable-next-line
          let func = new Function('data', col.formula)
          content = func([record])
        } catch (e) {
          console.warn(e)
          content = ''
        }
      } else {
        Object.keys(record).forEach(key => {
          let reg = new RegExp('@' + key + '@', 'ig')
          content = content.replace(reg, record[key])
        })
  
        if (col.eval !== 'false') {
          try {
            // eslint-disable-next-line
            content = eval(content)
          } catch (e) {
            console.info(content)
            console.warn(e)
            content = ''
          }
        }
      }

      content = content === undefined ? '' : content

      if (col.noValue === 'hide' && content === 0) {
        content = ''
      }

      if (col.round && typeof(content) === 'number') {
        content = Math.round(content * col.round) / col.round
        content = content.toFixed(col.decimal)
      }

      if (col.eval === 'func') {
        content = <span dangerouslySetInnerHTML={{__html: content}}></span>
      } else if (content !== '') {
        content = `${col.prefix || ''}${content}${col.postfix || ''}`
        content = content.replace(/\n/ig, '<br/>').replace(/\s/ig, '&nbsp;')
        content = <span dangerouslySetInnerHTML={{__html: content}}></span>
      }

      if (col.marks) {
        style = style ? {...style} : {}

        let mark = getMark(col.marks, record, style)

        if (mark.icon) {
          if (mark.position === 'front') {
            content = <span><MkIcon style={mark.innerStyle} type={mark.icon} /> {content}</span>
          } else {
            content = <span>{content} <MkIcon style={mark.innerStyle} type={mark.icon} /></span>
          }
        } else if (mark.innerStyle) {
          content = <span style={mark.innerStyle}>{content}</span>
        } else if (mark.space) {
          content = <><span dangerouslySetInnerHTML={{__html: mark.space}}></span>{content}</>
        } else if (mark.point) {
          if (mark.position === 'front') {
            content = <>{mark.point}{content}</>
          } else {
            content = <>{content}{mark.point}</>
          }
        }
      }

      children = content
    } else if (col.type === 'custom') {
      style.padding = '0px'
      if (col.style) {
        style = {...style, ...col.style}
      }

      children = (
        <CardCellComponent data={record} cards={col.config} elements={col.elements}/>
      )
    }

    return (<td className={className} style={style}>{children}</td>)
  }
}

class BodyAllCell extends React.Component {
  state = {}

  shouldComponentUpdate (nextProps, nextState) {
    return !is(fromJS(this.props.record), fromJS(nextProps.record)) || !is(fromJS(this.state), fromJS(nextState))
  }

  /**
   * @description 组件销毁，清除state更新，清除快捷键设置
   */
  componentWillUnmount () {
    this.setState = () => {
      return
    }
  }

  onColChange = (values) => {
    const { col, record } = this.props

    MKEmitter.emit('changeRecord' + col.tableId, {...record, ...values})
  }

  render() {
    let { col, record, style, className, ...resProps } = this.props

    if (!col) return (<td {...resProps} className={className} style={style}/>)

    let disabled = false
    let editable = false
    if (col.ctrlField) {
      disabled = col.ctrlValue.includes(record[col.ctrlField] + '')
    }

    let children = null
    if (col.type === 'text') {
      if (col.editable === 'true' && !disabled) {
        editable = true
        let _value = record[col.field] !== undefined ? record[col.field] : ''
        
        if (!col.editType || col.editType === 'text') {
          children = (
            <MkInput config={col} lineId={record.$$uuid} defaultValue={_value} autoFocus={false} onChange={this.onColChange}/>
          )
        } else if (col.editType === 'switch') {
          children = (
            <MkSwitch config={col} lineId={record.$$uuid} defaultValue={_value} autoFocus={false} onChange={this.onColChange}/>
          )
        } else if (col.editType === 'date') {
          children = (
            <MkDatePicker config={col} lineId={record.$$uuid} defaultValue={_value || null} autoFocus={false} onChange={this.onColChange}/>
          )
        } else if (col.editType === 'popSelect') {
          let showValue = ''
          if (col.showField) {
            showValue = record[col.showField] || ''
          }
          
          children = (
            <MKPopSelect config={col} lineId={record.$$uuid} defaultValue={_value} showValue={showValue} BID={record.$$BID} autoFocus={false} onChange={this.onColChange}/>
          )
        } else {
          children = (
            <MkSelect config={col} lineId={record.$$uuid} defaultValue={_value} autoFocus={false} onChange={this.onColChange} />
          )
        }
      } else {
        let content = ''
        if (record[col.field] !== undefined) {
          content = `${record[col.field]}`
        }

        if (col.editType === 'select' && col.options.length > 0) {
          content = col.map.get(content) || content
        } else if (col.editType === 'switch') {
          if (content === col.openVal) {
            content = col.openText
          } else if (content === col.closeVal) {
            content = col.closeText
          }
        } else if (col.editType === 'popSelect') {
          if (col.showField) {
            content = record[col.showField] || content
          }
        }

        if (content !== '') {
          if (col.textFormat === 'YYYY-MM-DD' && /^[1-9]\d{3}(-|\/)(0[1-9]|1[0-2])(-|\/)(0[1-9]|[1-2][0-9]|3[0-1])/.test(content)) {
            content = `${content.substr(0, 4)}-${content.substr(5, 2)}-${content.substr(8, 2)}`
          } else if (col.textFormat === 'YYYY-MM-DD HH:mm:ss' && /^[1-9]\d{3}(-|\/)(0[1-9]|1[0-2])(-|\/)(0[1-9]|[1-2][0-9]|3[0-1]).([0-1][0-9]|2[0-3]):[0-5][0-9]:[0-5][0-9]/.test(content)) {
            content = `${content.substr(0, 4)}-${content.substr(5, 2)}-${content.substr(8, 2)} ${content.substr(11, 2)}:${content.substr(14, 2)}:${content.substr(17, 2)}`
          } else if (col.textFormat === 'encryption') {
            content = <span>{col.prefix || ''}<Encrypts value={content} />{col.postfix || ''}</span>
          }

          if (col.noValue === 'hide' && content < '1949-10-02') {
            content = ''
          }

          if (col.textFormat !== 'encryption') {
            content = (col.prefix || '') + content + (col.postfix || '')
          }
        }

        if (col.marks) {
          style = style ? {...style} : {}

          let mark = getMark(col.marks, record, style)

          if (mark.icon) {
            if (mark.position === 'front') {
              content = <span><MkIcon style={mark.innerStyle} type={mark.icon} /> {content}</span>
            } else {
              content = <span>{content} <MkIcon style={mark.innerStyle} type={mark.icon} /></span>
            }
          } else if (mark.innerStyle) {
            content = <span style={mark.innerStyle}>{content}</span>
          } else if (mark.space) {
            content = <><span dangerouslySetInnerHTML={{__html: mark.space}}></span>{content}</>
          } else if (mark.point) {
            if (mark.position === 'front') {
              content = <>{mark.point}{content}</>
            } else {
              content = <>{content}{mark.point}</>
            }
          }
        }

        if (col.editable === 'true' && disabled) {
          content = <span style={{display: 'inline-block', padding: '0 6px'}}>{content}</span>
        }

        children = content
      }
    } else if (col.type === 'number') {
      if (col.editable === 'true' && !disabled) {
        editable = true
        let _value = record[col.field] !== undefined ? record[col.field] : ''

        if (col.noValue === 'hide' && _value === 0) {
          _value = ''
        }

        children = (
          <MkInputNumber config={col} lineId={record.$$uuid} defaultValue={_value} autoFocus={false} onChange={this.onColChange}/>
        )
      } else {
        let content = ''
        try {
          content = parseFloat(record[col.field])
          if (isNaN(content)) {
            content = ''
          }
          if (col.noValue === 'hide' && content === 0) {
            content = ''
          }
        } catch (e) {
          content = ''
        }

        if (content !== '') {
          if (col.round) {
            content = Math.round(content * col.round) / col.round
          }
          if (col.format === 'percent') {
            content = content * 100
          } else if (col.format === 'abs') {
            content = Math.abs(content)
          }
          if (col.round) {
            content = content.toFixed(col.decimal)
          }
    
          if (col.format === 'thdSeparator') {
            content = content + ''
            content = content.replace(/\d{1,3}(?=(\d{3})+(\.\d*)?$)/g, '$&,')
          }
    
          content = (col.prefix || '') + content + (col.postfix || '')
        }

        if (col.marks) {
          style = style ? {...style} : {}

          let mark = getMark(col.marks, record, style)

          if (mark.icon) {
            if (mark.position === 'front') {
              content = <span><MkIcon style={mark.innerStyle} type={mark.icon} /> {content}</span>
            } else {
              content = <span>{content} <MkIcon style={mark.innerStyle} type={mark.icon} /></span>
            }
          } else if (mark.innerStyle) {
            content = <span style={mark.innerStyle}>{content}</span>
          } else if (mark.space) {
            content = <><span dangerouslySetInnerHTML={{__html: mark.space}}></span>{content}</>
          } else if (mark.point) {
            if (mark.position === 'front') {
              content = <>{mark.point}{content}</>
            } else {
              content = <>{content}{mark.point}</>
            }
          }
        }

        if (col.editable === 'true' && disabled) {
          content = <span style={{display: 'inline-block', padding: '0 6px'}}>{content}</span>
        }

        children = content
      }
    } else if (col.type === 'textarea') {
      let content = ''
      if (record[col.field] !== undefined) {
        content = `${record[col.field]}`
      }

      if (content) {
        content = (col.prefix || '') + content + (col.postfix || '')
      }

      children = (
        <div>
          {content ? <Paragraph copyable ellipsis={{ rows: 3, expandable: true }}>{content}</Paragraph> : null }
        </div>
      )
    } else if (col.type === 'formula') {
      let content = col.formula

      if (col.eval === 'func') {
        try {
          // eslint-disable-next-line
          let func = new Function('data', col.formula)
          content = func([record])
        } catch (e) {
          console.warn(e)
          content = ''
        }
      } else {
        Object.keys(record).forEach(key => {
          let reg = new RegExp('@' + key + '@', 'ig')
          content = content.replace(reg, record[key])
        })
  
        if (col.eval !== 'false') {
          try {
            // eslint-disable-next-line
            content = eval(content)
          } catch (e) {
            console.info(content)
            console.warn(e)
            content = ''
          }
        }
      }

      content = content === undefined ? '' : content

      if (col.noValue === 'hide' && content === 0) {
        content = ''
      }

      if (col.round && typeof(content) === 'number') {
        content = Math.round(content * col.round) / col.round
        content = content.toFixed(col.decimal)
      }

      if (col.eval === 'func') {
        content = <span dangerouslySetInnerHTML={{__html: content}}></span>
      } else if (content !== '') {
        content = `${col.prefix || ''}${content}${col.postfix || ''}`
        content = content.replace(/\n/ig, '<br/>').replace(/\s/ig, '&nbsp;')
        content = <span dangerouslySetInnerHTML={{__html: content}}></span>
      }

      if (col.marks) {
        style = style ? {...style} : {}

        let mark = getMark(col.marks, record, style)

        if (mark.icon) {
          if (mark.position === 'front') {
            content = <span><MkIcon style={mark.innerStyle} type={mark.icon} /> {content}</span>
          } else {
            content = <span>{content} <MkIcon style={mark.innerStyle} type={mark.icon} /></span>
          }
        } else if (mark.innerStyle) {
          content = <span style={mark.innerStyle}>{content}</span>
        } else if (mark.space) {
          content = <><span dangerouslySetInnerHTML={{__html: mark.space}}></span>{content}</>
        } else if (mark.point) {
          if (mark.position === 'front') {
            content = <>{mark.point}{content}</>
          } else {
            content = <>{content}{mark.point}</>
          }
        }
      }

      children = content
    } else if (col.type === 'custom') {
      style.padding = '0px'
      if (col.style) {
        style = {...style, ...col.style}
      }

      children = (
        <CardCellComponent data={record} cards={col.config} elements={col.elements}/>
      )
    }

    if (editable) {
      return (<td onClick={(e) => e.stopPropagation()} className={'editing_all_table_cell ' + className} style={style}>{children}</td>)
    }

    return (<td className={'editing_all_table_cell ' + className} style={style}>{children}</td>)
  }
}

class NormalTable extends Component {
  static propTpyes = {
    MenuID: PropTypes.string,        // 菜单Id
    setting: PropTypes.object,       // 表格全局设置：tableType（表格是否可选、单选、多选）、actionfixed（按钮固定）
    columns: PropTypes.array,        // 表格列
    lineMarks: PropTypes.any,        // 行标记
    fields: PropTypes.array,         // 组件字段集
    BID: PropTypes.any,              // 主表ID
    total: PropTypes.any,            // 总数
    loading: PropTypes.bool,         // 表格加载中
    refreshdata: PropTypes.func,     // 表格中排序列、页码的变化时刷新
    chgSelectData: PropTypes.func,
    allSearch: PropTypes.any,
    colsCtrls: PropTypes.any
  }

  state = {
    edData: [],
    selectedRowKeys: [],  // 表格中选中行
    tableId: '',          // 表格ID
    pageIndex: 1,         // 初始页面索引
    pageSize: 10,         // 每页数据条数
    columns: null,        // 显示列
    forms: [],
    orderfields: {},      // 排序id与field转换
    loading: false,
    pageOptions: [],
    deForms: null,
    visible: false,
    midData: null,
    allColumns: null,
    checkForms: [],
    allForms: [],
    reseting: false
  }

  timer = null
  focusId = ''
  blurId = ''
  colId = ''

  UNSAFE_componentWillMount () {
    const { setting, fields, columns, BID, colsCtrls } = this.props
    let orderfields = {}

    let _columns = []
    let deForms = []
    let _forms = {}
    let hasBid = false
    let checkForms = []
    let allForms = []

    let getColumns = (cols, sk) => {
      return cols.map(item => {
        let cell = null
  
        if (item.type === 'colspan') {
          cell = { title: item.label, align: item.Align, $key: item.uuid }
          cell.children = getColumns(item.subcols, sk || item.uuid)
        } else {
          if (item.editable === 'true') {
            _forms[item.field] = item

            allForms.push({uuid: sk || item.uuid, field: item.field})
            checkForms.push(item.field)

            if (item.type === 'text' && item.editType === 'select') {
              if (item.resourceType === '1') {
                let _option = Utils.getSelectQueryOptions(item)
  
                if (/@BID@/ig.test(_option.sql) && setting.supModule) {
                  hasBid = true
                }
        
                item.base_sql = _option.sql
                item.arr_field = _option.field
        
                deForms.push(item)
              }
            }
          }
    
          if (item.field) {
            orderfields[item.uuid] = item.field
          } else if (item.sortField) {
            orderfields[item.uuid] = item.sortField
          }

          cell = {
            align: item.Align,
            dataIndex: item.uuid,
            title: item.editable === 'true' ? <span>{item.label}<EditOutlined className="system-color mk-edit-sign"/></span> : item.label,
            sorter: (item.field || item.sortField) && item.IsSort === 'true',
            width: item.Width || 120,
            $key: item.uuid,
            onCell: record => ({
              record,
              col: item
            })
          }
        }
  
        return cell
      })
    }

    _columns = getColumns(columns)

    let forms = []
    fields.forEach(item => {
      if (item.field === setting.primaryKey) return

      if (_forms[item.field]) {
        let _item = {..._forms[item.field]}
        if (_item.editType === 'date') {
          _item.datatype = _item.declareType || 'datetime'
        } else {
          _item.datatype = item.datatype
        }

        forms.push(_item)
      } else {
        forms.push({...item, $sort: 999})
      }
    })

    forms.sort((a, b) => a.$sort - b.$sort)

    let size = (setting.pageSize || 10) + ''
    let pageOptions = ['10', '25', '50', '100', '500', '1000']

    if (!pageOptions.includes(size)) {
      pageOptions.push(size)
      pageOptions = pageOptions.sort((a, b) => a - b)
    }

    if (setting.maxPageSize) {
      pageOptions = pageOptions.filter(item => item <= setting.maxPageSize)
    }

    let allColumns = null
    if (colsCtrls) {
      allColumns = [..._columns]
      checkForms = []
      _columns = this.getCurColumns(_columns, this.props.allSearch, allForms, checkForms)
    } else {
      allForms = null
    }

    checkForms = Array.from(new Set(checkForms))

    this.setState({
      forms,
      allForms,
      allColumns,
      checkForms,
      pageSize: setting.pageSize || 10,
      pageOptions,
      columns: _columns,
      tableId: setting.tableId,
      orderfields,
      deForms: hasBid ? deForms : null
    }, () => {
      if (deForms.length > 0 && (!hasBid || BID)) {
        this.improveActionForm(deForms, BID)
      }
    })
  }

  shouldComponentUpdate (nextProps, nextState) {
    return !is(fromJS(this.props), fromJS(nextProps)) || !is(fromJS(this.state), fromJS(nextState))
  }

  UNSAFE_componentWillReceiveProps(nextProps) {
    const { BID, parCtrl } = this.props
    const { deForms } = this.state

    if (deForms && nextProps.BID !== BID) {
      this.improveActionForm(deForms, nextProps.BID)
    }
    if (parCtrl && !is(fromJS(this.props.columns), fromJS(nextProps.columns))) {
      let getColumns = (cols, sk) => {
        return cols.map(item => {
          let cell = null
    
          if (item.type === 'colspan') {
            cell = { title: item.label, align: item.Align, $key: item.uuid }
            cell.children = getColumns(item.subcols, sk || item.uuid)
          } else {
            cell = {
              align: item.Align,
              dataIndex: item.uuid,
              title: item.editable === 'true' ? <span>{item.label}<EditOutlined className="system-color mk-edit-sign"/></span> : item.label,
              sorter: (item.field || item.sortField) && item.IsSort === 'true',
              width: item.Width || 120,
              $key: item.uuid,
              onCell: record => ({
                record,
                col: item
              })
            }
          }
    
          return cell
        })
      }
  
      let _columns = getColumns(nextProps.columns)

      this.setState({
        columns: _columns
      })
    }
  }

  componentDidMount () {
    const { setting } = this.props
    const { tableId } = this.state
    
    if (setting.commit === 'change') {
      MKEmitter.addListener('colBlur' + tableId, this.colBlur)
      MKEmitter.addListener('colFocus' + tableId, this.colFocus)
    }

    MKEmitter.addListener('resetTable', this.resetTable)
    MKEmitter.addListener('nextLine' + tableId, this.nextLine)
    MKEmitter.addListener('addRecord' + tableId, this.plusLine)
    MKEmitter.addListener('delRecord' + tableId, this.delRecord)
    MKEmitter.addListener('transferData' + tableId, this.transferData)
    MKEmitter.addListener('changeRecord' + tableId, this.changeRecord)
  }

  /**
   * @description 组件销毁，清除state更新
   */
  componentWillUnmount () {
    const { tableId } = this.state

    this.setState = () => {
      return
    }

    MKEmitter.removeListener('resetTable', this.resetTable)
    MKEmitter.removeListener('colBlur' + tableId, this.colBlur)
    MKEmitter.removeListener('colFocus' + tableId, this.colFocus)
    MKEmitter.removeListener('nextLine' + tableId, this.nextLine)
    MKEmitter.removeListener('addRecord' + tableId, this.plusLine)
    MKEmitter.removeListener('delRecord' + tableId, this.delRecord)
    MKEmitter.removeListener('transferData' + tableId, this.transferData)
    MKEmitter.removeListener('changeRecord' + tableId, this.changeRecord)
  }

  colBlur = (lineId, colId, defer) => {
    if (defer && this.focusId === lineId && this.colId !== colId) {
      return
    }
    
    this.blurId = lineId
    this.focusId = ''

    this.timer && clearTimeout(this.timer)

    this.timer = setTimeout(() => {
      if (!this.focusId || this.focusId !== this.blurId) {
        this.checkLine()
      }
    }, 150)
  }

  colFocus = (lineId, colId) => {
    this.focusId = lineId
    this.colId = colId
  }

  checkLine = () => {
    const { edData, forms, checkForms } = this.state

    let data = edData.filter(item => item.$$uuid === this.blurId)[0]

    if (!data) return

    let record = fromJS(data).toJS()

    let value = ''
    Object.keys(record).sort().forEach(key => {
      if (/^\$/.test(key)) return
      value += record[key]
    })

    if (record.$sign === md5(value)) return

    let err = ''
    forms.forEach(col => {
      let check = true
      if (col.editable !== 'true' || !checkForms.includes(col.field)) {
        check = false
      }
      if (col.ctrlField && col.ctrlValue.includes(record[col.ctrlField] + '')) {
        check = false
      }

      if (!check) {
        if (col.type === 'number') {
          record[col.field] = +record[col.field]
          if (isNaN(record[col.field])) {
            record[col.field] = 0
          }
        } else {
          record[col.field] = record[col.field] !== undefined ? (record[col.field] + '') : ''
        }
        return
      }

      if (err) return

      if (col.type === 'text') {
        let val = record[col.field] !== undefined ? (record[col.field] + '') : ''
        if (col.required === 'true' && !val) {
          err = `${col.label}不可为空`
        } else if (col.datatype === 'datetime' && !val) {
          val = '1949-10-01'
        }
        record[col.field] = val
      } else if (col.type === 'number') {
        let val = record[col.field]

        if (col.required === 'true' && !val) {
          err = `${col.label}不可为${col.noValue === 'hide' ? '空' : '0'}`
        } else if (col.noValue === 'hide' && !val) {
          if (col.clearField && checkForms.includes(col.clearField) && !record[col.clearField]) {
            err = `请填写 ${col.label} 或 ${col.clearName}`
          }
          val = 0
        } else if (!val && val !== 0) {
          err = `${col.label}不可为空`
        } else {
          val = +val
          if (isNaN(val)) {
            err = `${col.label}数据格式错误`
            return
          }
  
          val = +val.toFixed(col.decimal || 0)
          
          if (typeof(col.max) === 'number' && val > col.max) {
            err = `${col.label}不可大于${col.max}`
          } else if (typeof(col.min) === 'number' && val < col.min) {
            err = `${col.label}不可小于${col.min}`
          }
        }

        record[col.field] = val
      }
    })

    if (err) {
      message.warning(err)
      return
    }

    this.submit(record)
  }

  getCurColumns = (columns, allSearch, allForms, checkForms) => {
    const { colsCtrls } = this.props

    let values = {}
    allSearch.forEach(item => {
      values[item.key] = item.value
    })
    let cols = null
    colsCtrls.some(item => {
      let originVal = item.field.map(f => values[f] || '').join('')
      let contrastVal = item.contrastValue
      let result = false

      if (item.match === '=') {
        result = originVal === contrastVal
      } else if (item.match === '!=') {
        result = originVal !== contrastVal
      } else if (item.match === 'regexp') {
        let reg = new RegExp(item.contrastValue, 'ig')
        result = reg.test(originVal)
      } else {
        originVal = isNaN(originVal) ? originVal : +originVal
        contrastVal = isNaN(contrastVal) ? contrastVal : +contrastVal
        if (item.match === '>') {
          result = originVal > contrastVal
        } else if (item.match === '<') {
          result = originVal < contrastVal
        }
      }

      if (!result) return false

      cols = item.cols

      return true
    })

    if (cols) {
      allForms.forEach(item => {
        if (cols.includes(item.uuid)) {
          checkForms.push(item.field)
        }
      })

      return columns.filter(col => cols.includes(col.$key))
    }

    allForms.forEach(item => {
      checkForms.push(item.field)
    })

    return columns
  }

  transferData = (data, type) => {
    const { edData, tableId } = this.state

    if (type === 'clear') {
      this.setState({edData: [], midData: []})
      return
    } else if (type === 'delete') {

    } else if (type === 'line') {
      let value = ''
      Object.keys(data).sort().forEach(key => {
        if (/^\$/.test(key)) return
        value += data[key]
      })
      data.$sign = md5(value)
    } else {
      data.forEach(item => {
        let value = ''
        Object.keys(item).sort().forEach(key => {
          if (/^\$/.test(key)) return
          value += item[key]
        })
        item.$sign = md5(value)
      })
    }

    if (type === 'delete') {
      let _edData = this.state.edData.filter(item => item.$$uuid !== data)

      this.setState({edData: _edData, reseting: true}, () => {
        this.setState({reseting: false})

        if (this.focusId) {
          setTimeout(() => {
            MKEmitter.emit('setFocus' + tableId, this.focusId, this.colId)
          }, 10)
        }
      })
    } else if (type === 'line') {
      let _edData = this.state.edData.map(item => {
        if (item.$$uuid === data.$$uuid) {
          return data
        } else {
          return item
        }
      })

      this.setState({edData: _edData, reseting: true}, () => {
        this.setState({reseting: false})

        if (this.focusId) {
          setTimeout(() => {
            MKEmitter.emit('setFocus' + tableId, this.focusId, this.colId)
          }, 10)
        }
      })
    } else {
      let index = edData.findIndex(item => !item.$origin && !item.$forbid)

      if (index > -1) {
        this.setState({visible: true, midData: data})
      } else {
        this.updateMutil(data)
      }
    }
  }

  updateMutil = (data) => {
    const { setting, colsCtrls, allSearch } = this.props
    const { allColumns, allForms } = this.state

    if (colsCtrls) {
      let checkForms = []
      let columns = this.getCurColumns(allColumns, allSearch, allForms, checkForms)

      checkForms = Array.from(new Set(checkForms))

      this.setState({
        checkForms,
        columns: columns,
        reseting: true,
        edData: data,
        visible: false,
        midData: null
      }, () => {
        this.setState({
          reseting: false
        })
      })

      return
    }

    if (setting.editType === 'multi' && data.length > 0) {
      this.setState({edData: data, visible: false, midData: null, reseting: true}, () => {
        this.setState({reseting: false})
      })
    } else {
      this.setState({edData: data, visible: false, midData: null})
    }

    if (setting.addable && data.length === 0) {
      setTimeout(() => {
        this.plusLine()
      }, 10)
    }
  }

  improveActionForm = (deForms, BID) => {
    let deffers = []
    let mainItems = []  // 云端或单点数据
    let localItems = [] // 本地数据
    let debug = window.GLOB.debugger === true
    let _sql = `Declare @mk_departmentcode nvarchar(512),@mk_organization nvarchar(512),@mk_user_type nvarchar(20)  select @mk_departmentcode='${sessionStorage.getItem('departmentcode') || ''}',@mk_organization='${sessionStorage.getItem('organization') || ''}',@mk_user_type='${sessionStorage.getItem('mk_user_type') || ''}'\n`
    let _sso = _sql

    deForms.forEach(item => {
      if (item.database === 'sso') {
        let sql = _sso + item.base_sql
        _sso = ''

        sql = sql.replace(/@BID@/ig, `'${BID}'`)

        if (debug) {
          console.info(sql)
        }

        sql = sql.replace(/%/ig, ' mpercent ')

        mainItems.push(`select '${item.uuid}' as obj_name,'${item.arr_field}' as arr_field,'${window.btoa(window.encodeURIComponent(sql))}' as LText`)
      } else {
        let sql = _sql + item.base_sql
        _sql = ''

        sql = sql.replace(/@BID@/ig, `'${BID}'`)

        if (debug) {
          console.info(sql)
        }

        sql = sql.replace(/%/ig, ' mpercent ')
        
        localItems.push(`select '${item.uuid}' as obj_name,'${item.arr_field}' as arr_field,'${window.btoa(window.encodeURIComponent(sql))}' as LText`)
      }
    })

    // 本地请求
    let param = {
      func: 'sPC_Get_SelectedList',
      LText: localItems.join(' union all '),
      obj_name: '',
      arr_field: '',
      table_type: 'Y',
      BID: BID || ''
    }

    if (param.LText) {
      if (window.GLOB.execType === 'x') {
        param.exec_type = 'x'
      }
      
      param.LText = Utils.formatOptions(param.LText, param.exec_type)
      param.timestamp = moment().format('YYYY-MM-DD HH:mm:ss')
      param.secretkey = Utils.encrypt(window.GLOB.execType === 'x' ? '' : param.LText, param.timestamp)

      deffers.push(
        new Promise(resolve => {
          Api.getSystemCacheConfig(param, false).then(res => {
            if (!res.status) {
              notification.warning({
                top: 92,
                message: res.message,
                duration: 5
              })
            }
            resolve(res)
          })
        })
      )
    }

    // 系统请求
    let mainparam = {
      func: 'sPC_Get_SelectedList',
      LText: mainItems.join(' union all '),
      obj_name: '',
      arr_field: '',
      table_type: 'Y',
      BID: BID || ''
    }

    if (mainparam.LText) {
      if (window.GLOB.execType === 'x') {
        mainparam.exec_type = 'x'
      }

      mainparam.LText = Utils.formatOptions(mainparam.LText, mainparam.exec_type)
      mainparam.timestamp = moment().format('YYYY-MM-DD HH:mm:ss')
      mainparam.secretkey = Utils.encrypt(window.GLOB.execType === 'x' ? '' : mainparam.LText, mainparam.timestamp)

      if (window.GLOB.mainSystemApi) {
        mainparam.rduri = window.GLOB.mainSystemApi
      }

      deffers.push(
        new Promise(resolve => {
          Api.getSystemCacheConfig(mainparam, false).then(res => {
            if (!res.status) {
              notification.warning({
                top: 92,
                message: res.message,
                duration: 5
              })
            }
            resolve(res)
          })
        })
      )
    }

    Promise.all(deffers).then(response => {
      let result = {...response[0], ...(response[1] || {})}

      delete result.ErrCode
      delete result.ErrMesg
      delete result.message
      delete result.status

      this.resetFormList(result)
    })
  }

  resetFormList = (result) => {
    const { columns, edData } = this.state

    let _edColumns = []

    let reCols = {}
    this.props.columns.forEach(item => {
      if (item.resourceType === '1' && result[item.uuid] && result[item.uuid].length > 0) {
        let options = []
        let all = false
        result[item.uuid].forEach(cell => {
          let _cell = {key: Utils.getuuid()}

          _cell.value = cell[item.valueField]
          _cell.label = cell[item.valueText]

          if (!_cell.label && _cell.label !== 0) {
            if (!all) {
              _cell.label = '全部'
              all = true
            } else {
              return
            }
          }

          if (item.map.has(_cell.value)) return
          item.map.set(_cell.value, _cell.label)

          if (item.linkSubField) {
            item.linkSubField.forEach(m => {
              _cell[m] = cell[m] === undefined ? '' : cell[m]
            })
          }

          if (item.disableField && cell[item.disableField] && /^true$/ig.test(cell[item.disableField])) {
            _cell.$disabled = true
          }
  
          options.push(_cell)
        })

        item.options = options

        reCols[item.uuid] = fromJS(item).toJS()
      }
    })

    _edColumns = columns.map(item => {
      if (reCols[item.dataIndex]) {
        item.onCell = record => ({
          record,
          col: reCols[item.dataIndex]
        })
      }

      return item
    })

    let _cols = this.state.allColumns
    if (_cols) {
      _cols = _cols.map(item => {
        if (reCols[item.dataIndex]) {
          item.onCell = record => ({
            record,
            col: reCols[item.dataIndex]
          })
        }
  
        return item
      })
    }

    this.setState({columns: [], edData: [], allColumns: _cols}, () => {
      this.setState({columns: _edColumns, edData: edData})
    })
  }
  
  nextLine = (lineId, colId) => {
    const { setting } = this.props
    const { edData, tableId } = this.state

    let index = edData.findIndex(item => item.$$uuid === lineId)
    let next = edData[index + 1] || null
    
    if (next) {
      MKEmitter.emit('setFocus' + tableId, next.$$uuid, colId)
    } else if (setting.addable) {
      setTimeout(() => {
        this.plusLine(colId)
      }, 10)
    } else if (edData[index] && (setting.commit === 'all' || setting.commit === 'amend')) {
      setTimeout(() => {
        this.submit()
      }, 10)
    }
  }

  plusLine = (colId, lineId) => {
    const { BID } = this.props
    const { forms, tableId } = this.state

    let edData = fromJS(this.state.edData).toJS()
    let item = edData.length > 0 ? {...edData[edData.length - 1]} : {}
    let index = null
    
    if (lineId) {
      index = edData.findIndex(item => lineId === item.$$uuid)
      index = index === -1 ? null : index
      
      if (index !== null) {
        item = {...edData[index]}
      }
    }

    item.key = edData.length
    item.$$uuid = Utils.getguid()
    item.$type = 'add'
    item.$forbid = true
    item.$Index = ''
    item.$$BID = BID || ''

    forms.forEach(col => {
      if (col.initval !== '$copy') {
        item[col.field] = col.initval
      }
      if (col.type === 'number') {
        item[col.field] = +item[col.field]
        if (isNaN(item[col.field])) {
          item[col.field] = 0
        }
      }
      if (item[col.field] === undefined) {
        item[col.field] = ''
      }
    })

    let value = ''
    Object.keys(item).sort().forEach(key => {
      if (/^\$/.test(key)) return
      value += item[key]
    })
    item.$sign = md5(value)

    if (index === null) {
      edData.push(item)
    } else {
      edData.splice(index, 0, item)
    }

    this.setState({edData: edData}, () => {
      if (colId) {
        MKEmitter.emit('setFocus' + tableId, item.$$uuid, colId)
      }
    })
  }

  delRecord = (record) => {
    const { setting } = this.props
    const { edData } = this.state

    if (record.$type === 'add') {
      let _data = edData.filter(item => item.$$uuid !== record.$$uuid)
      this.setState({edData: _data})
    } else {
      let _data = fromJS(edData).toJS().map(item => {
        if (item.$$uuid === record.$$uuid) {
          item.$deleted = true
          item.$origin = false
          item.$type = 'del'

          record.$deleted = true
          record.$origin = false
          record.$type = 'del'
        }

        return item
      })

      this.setState({edData: _data}, () => {
        if (setting.commit === 'change') {
          this.submit(record)
        }
      })
    }
  }

  changeRecord = (record) => {
    const { setting } = this.props
    const { edData } = this.state

    let data = edData.filter(item => item.$$uuid === record.$$uuid)[0]

    if (!data) return
    
    if (is(fromJS(data), fromJS(record))) return

    delete record.$forbid

    let value = ''
    Object.keys(record).sort().forEach(key => {
      if (/^\$/.test(key)) return
      value += record[key]
    })

    let sign = md5(value)

    if (record.$sign === sign) {
      record.$origin = true
      record.$lock = false
    } else {
      record.$origin = false
      record.$lock = true
    }

    let _data = edData.map(item => {
      if (item.$$uuid === record.$$uuid) {
        return record
      } else {
        return item
      }
    })

    this.setState({edData: _data}, () => {
      if (setting.tableType && setting.hasAction && this.state.selectedRowKeys.includes(record.$$uuid)) {
        this.selectdata(this.state.selectedRowKeys)
      }
    })
  }

  checkData = () => {
    const { setting } = this.props
    const { edData, forms, checkForms, selectedRowKeys } = this.state

    let data = fromJS(edData).toJS()

    data = data.filter(item => !item.$forbid)

    if (setting.commit === 'amend') {
      data = data.filter(item => !item.$origin)
    } else if (setting.commit === 'check') {
      data = data.filter(item => selectedRowKeys.includes(item.$$uuid))

      if (data.length === 0) {
        notification.warning({
          top: 92,
          message: '请选择需要提交的数据！',
          duration: 5
        })
        return null
      }
    }

    if (data.length === 0) {
      notification.warning({
        top: 92,
        message: '数据未修改，不可提交！',
        duration: 5
      })
      return null
    }

    let err = ''
    let Index = 1
    data = data.map(item => {
      let line = []
      forms.forEach(col => {
        let check = true
        if (col.editable !== 'true' || item.$deleted || !checkForms.includes(col.field)) {
          check = false
        }
        if (col.ctrlField && col.ctrlValue.includes(item[col.ctrlField] + '')) {
          check = false
        }

        if (!check) {
          if (col.type === 'number') {
            item[col.field] = +item[col.field]
            if (isNaN(item[col.field])) {
              item[col.field] = 0
            }
          } else {
            item[col.field] = item[col.field] !== undefined ? (item[col.field] + '') : ''
          }
          return
        }

        if (col.type === 'text') {
          let val = item[col.field] !== undefined ? (item[col.field] + '') : ''
          if (col.required === 'true' && !val) {
            line.push(`${col.label}不可为空`)
          } else if (col.datatype === 'datetime' && !val) {
            val = '1949-10-01'
          }
          item[col.field] = val
        } else if (col.type === 'number') {
          let val = item[col.field]
          if (col.required === 'true' && !val) {
            err = `${col.label}不可为${col.noValue === 'hide' ? '空' : '0'}`
          } else if (col.noValue === 'hide' && !val) {
            if (col.clearField && checkForms.includes(col.clearField) && !item[col.clearField]) {
              let msg = `请填写 ${col.label} 或 ${col.clearName}`
              if (!line.includes(msg)) {
                line.push(msg)
              }
            }
            val = 0
          } else if (!val && val !== 0) {
            line.push(`${col.label}不可为空`)
            return
          } else {
            val = +val
            if (isNaN(val)) {
              line.push(`${col.label}数据格式错误`)
              return
            }
  
            val = +val.toFixed(col.decimal || 0)
            
            if (typeof(col.max) === 'number' && val > col.max) {
              line.push(`${col.label}不可大于${col.max}`)
            } else if (typeof(col.min) === 'number' && val < col.min) {
              line.push(`${col.label}不可小于${col.min}`)
            }
          }

          item[col.field] = val
        }
      })

      if (line.length > 0) {
        err += `第${Index}行：` + line.join('，') + '；'
      }
      if (!item.$deleted) {
        Index++
      }

      return item
    })

    if (err) {
      notification.warning({
        top: 92,
        message: err,
        duration: 5
      })

      return null
    }

    return data
  }

  submit = (record) => {
    const { submit, BID, setting } = this.props
    const { forms } = this.state

    this.setState({visible: false, midData: null})

    if (setting.supModule && !BID) {
      notification.warning({
        top: 92,
        message: '需要上级主键值！',
        duration: 5
      })

      return
    }

    let data = null
    if (record) {
      data = [record]
    } else {
      data = this.checkData()
    }

    if (!data) return

    let result = getEditTableSql(submit, data, forms)

    let param = {
      excel_in: result.lines,
      BID: BID || ''
    }

    this.setState({
      loading: true
    })

    if (submit.intertype === 'system') { // 系统存储过程
      param.func = 'sPC_TableData_InUpDe'
      
      delete param.excel_in

      param.exec_type = window.GLOB.execType || 'y'
      param.LText = Utils.formatOptions(result.sql, param.exec_type)
      param.timestamp = moment().format('YYYY-MM-DD HH:mm:ss')
      param.secretkey = Utils.encrypt('', param.timestamp)

      param.menuname = submit.logLabel

      if (window.GLOB.probation) {
        param.s_debug_type = 'Y'
      }

      Api.genericInterface(param).then((res) => {
        if (res.status) {
          this.execSuccess(res, record)
        } else {
          this.execError(res, record)
        }
      }, (error) => {
        if (error && error.ErrCode === 'LoginError') return

        this.execError({})
      })
    } else if (submit.intertype === 'inner' && submit.innerFunc) { // 自定义存储过程
      param.func = submit.innerFunc

      if (submit.recordUser === 'true') {
        param.username = sessionStorage.getItem('User_Name') || ''
        param.fullname = sessionStorage.getItem('Full_Name') || ''
      }

      Api.genericInterface(param).then((res) => {
        if (res.status) {
          this.execSuccess(res, record)
        } else {
          this.execError(res, record)
        }
      }, (error) => {
        if (error && error.ErrCode === 'LoginError') return

        this.execError({})
      })
    }
  }

  execSuccess = (res, record) => {
    const { submit } = this.props
    const { edData } = this.state

    if (res && res.ErrCode === 'S') { // 执行成功
      notification.success({
        top: 92,
        message: res.message || '执行成功',
        duration: submit.stime ? submit.stime : 2
      })
    } else if (res && res.ErrCode === 'Y') { // 执行成功
      Modal.success({
        title: res.message || '执行成功'
      })
    } else if (res && res.ErrCode === '-1') { // 完成后不提示

    }

    let _edData = fromJS(edData).toJS()

    _edData = _edData.filter(item => {
      if (item.$deleted) return false

      if (!item.$forbid) {
        item.$type = 'upt'
        item.$origin = true
        item.$lock = false
      }

      let value = ''
      Object.keys(item).sort().forEach(key => {
        if (/^\$/.test(key)) return
        value += item[key]
      })

      item.$sign = md5(value)

      return true
    })

    this.setState({
      loading: false,
      edData: _edData
    })

    if (submit.closetab === 'true') {
      MKEmitter.emit('popclose')
    }

    if (submit.execSuccess !== 'never') {
      MKEmitter.emit('refreshByButtonResult', submit.$menuId, submit.execSuccess, submit, '', record ? [record] : null)
    }

    submit.syncComponentId && MKEmitter.emit('reloadData', submit.syncComponentId)
  }

  execError = (res, record) => {
    const { submit } = this.props

    if (res.ErrCode === 'E') {
      Modal.error({
        title: res.message || '执行失败！',
      })
    } else if (res.ErrCode === 'N') {
      notification.error({
        top: 92,
        message: res.message || '执行失败！',
        duration: submit.ntime ? submit.ntime : 10
      })
    } else if (res.ErrCode === 'F') {
      notification.error({
        className: 'notification-custom-error',
        top: 92,
        message: res.message || '执行失败！',
        duration: submit.ftime ? submit.ftime : 10
      })
    } else if (res.ErrCode === 'NM') {
      message.error(res.message || '执行失败！')
    }
    
    this.setState({
      loading: false
    })

    if (submit.execError !== 'never') {
      MKEmitter.emit('refreshByButtonResult', submit.$menuId, submit.execError, submit, '', record ? [record] : null)
    }
  }

  /**
   * @description 选中行
   */
  onSelectChange = selectedRowKeys => {
    this.setState({ selectedRowKeys })

    let activeId = ''
    if (selectedRowKeys.length > 0) {
      activeId = selectedRowKeys.slice(-1)[0]
    }
    this.changedata(activeId)
    this.selectdata(selectedRowKeys)
  }

  /**
   * @description 点击整行，触发切换， 判断是否可选，单选或多选，进行对应操作
   */
  changeRow = (id) => {
    const { setting } = this.props

    if (!setting.tableType) return
    
    let newkeys = fromJS(this.state.selectedRowKeys).toJS()

    let activeId = ''
    if (setting.tableType === 'radio') {
      activeId = id
      newkeys = [id]
      this.setState({ selectedRowKeys: newkeys })
    } else {
      if (newkeys.includes(id)) {
        newkeys = newkeys.filter(item => item !== id)

        if (newkeys.length > 0) {
          activeId = newkeys.slice(-1)[0]
        }
      } else {
        activeId = id
        newkeys.push(id)
      }

      this.setState({ selectedRowKeys: newkeys })
    }

    this.changedata(activeId)
    this.selectdata(newkeys)
  }

  changedata = (id) => {
    const { MenuID } = this.props
    const { edData } = this.state

    let _data = ''

    if (id) {
      _data = edData.filter(item => item.$$uuid === id)[0] || ''
    }

    MKEmitter.emit('resetSelectLine', MenuID, id, _data)
  }

  selectdata = (keys) => {
    const { edData } = this.state

    let _data = edData.filter(item => keys.includes(item.$$uuid))

    this.props.chgSelectData(_data)
  }

  changeTable = (pagination, filters, sorter) => {
    const { orderfields } = this.state

    this.setState({
      pageIndex: pagination.current,
      pageSize: pagination.pageSize,
      selectedRowKeys: [],
    })

    sorter.field = orderfields[sorter.field] || ''

    this.props.refreshdata(pagination, filters, sorter)
  }

  resetTable = (id, repage) => {
    const { MenuID } = this.props

    if (id !== MenuID) return

    if (repage === 'false') {
      this.setState({
        selectedRowKeys: [],
      })
    } else {
      this.setState({
        pageIndex: 1,
        selectedRowKeys: [],
      })
    }
  }

  render() {
    const { setting, lineMarks, submit } = this.props
    const { edData, columns, loading, pageOptions, selectedRowKeys, visible, midData, reseting } = this.state

    if (reseting) return null

    const components = {
      body: {
        row: BodyRow,
        cell: setting.editType !== 'multi' ? BodyCell : BodyAllCell
      }
    }

    let _data = edData.filter(item => !item.$deleted)

    // 设置表格选择属性：单选、多选、不可选
    let rowSelection = null
    if (setting.tableType) {
      rowSelection = {
        selectedRowKeys,
        type: (setting.tableType === 'radio') ? 'radio' : 'checkbox',
        onChange: this.onSelectChange
      }
    }

    let _pagination = false
    if (setting.laypage) {
      _pagination = {
        current: this.state.pageIndex,
        pageSize: this.state.pageSize,
        pageSizeOptions: pageOptions,
        showSizeChanger: true,
        total: this.props.total || 0,
        showTotal: (total, range) => `${range[0]}-${range[1]} 共 ${total} 条`
      }
    }

    let height = setting.height || false
    if (height && height <= 100) {
      height = height + 'vh'
    }

    let style = {
      '--mk-table-border-color': setting.borderColor || '#e8e8e8',
      '--mk-table-color': setting.color || 'rgba(0, 0, 0, 0.65)',
      '--mk-table-font-size': setting.fontSize || '14px',
      '--mk-table-font-weight': setting.fontWeight || 'normal'
    }

    return (
      <>
        {setting.hasSubmit && edData.length > 0 ? <div className="edit-custom-table-btn-wrap" style={submit.wrapStyle}>
          <Button style={submit.style} onClick={() => setTimeout(() => {this.submit()}, 10)} loading={loading} className="submit-table" type="link">提交</Button>
        </div> : null}
        <div className={`edit-custom-table ${setting.tableHeader || ''} ${setting.parity === 'true' ? 'mk-parity' : ''} ${height ? 'fixed-height' : ''} ${setting.mode || ''} table-vertical-${setting.vertical || ''} mk-edit-${setting.editType || 'simple'}`} style={style}>
          <Table
            rowKey="$$uuid"
            components={components}
            size={setting.size || 'middle'}
            bordered={setting.bordered !== 'false'}
            rowSelection={rowSelection}
            columns={columns}
            dataSource={_data}
            loading={this.props.loading}
            scroll={{ x: '100%', y: height }}
            onRow={(record, index) => {
              return {
                lineMarks,
                title: setting.tipField ? record[setting.tipField] : '',
                data: record,
                onClick: () => {this.changeRow(record.$$uuid)},
              }
            }}
            onChange={this.changeTable}
            pagination={_pagination}
          />
          {setting.hasSubmit && _data.length > 10 ? <Button style={submit.style} onClick={() => setTimeout(() => {this.submit()}, 10)} loading={loading} className="submit-footer-table" type="link">提交</Button> : null}
        </div>
        <Modal
          className="mk-user-confirm"
          visible={visible}
          width={450}
          maskClosable={false}
          closable={false}
          footer={[
            <Button key="cancel" onClick={() => { this.setState({ visible: false, midData: null }) }}>取消</Button>,
            <Button key="refresh" className="table-refresh" onClick={() => { midData && this.updateMutil(midData) }}>刷新表格</Button>,
            <Button key="confirm" type="primary" onClick={() => setTimeout(() => {this.submit()}, 10)}>提交数据</Button>
          ]}
          destroyOnClose
        >
          <div><QuestionCircleOutlined />表格中有数据尚未提交</div>
        </Modal>
      </>
    )
  }
}

export default NormalTable